# -*- coding: utf-8 -*-
"""
Created on Fri Aug  5 14:47:38 2016

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import super
"""
from future import standard_library
standard_library.install_aliases()
"""
from .form import form
from ..transform.graphTransform import graphTransform, graphTransform3D

"""
 * A multiply-form denotes a form of a multiply-associated sort.
 * Multiply-forms may contain any number of individuals.
 * Different types of multiply-forms are distinguished based on the operational
 * behavior of their forms. 
 *
 * A {DiscreteForm} specifies a discrete behavior corresponding to a set: any 
 * two individuals are either identical or disjoint.
 * An {IntervalForm} specifies a behavior of intervals on a
 * one-dimensional axis: if two intervals overlap, these are combined into a
 * single interval.
 * An {ArealForm} applies to two-dimensional planar segments and
 * three-dimensional volumes, and uses a boundary classification approach to
 * determine overlapping segment parts.
 * A {RelationalForm} applies to (bi-directional) relations. While its
 * behavior is that of a discrete form, relations are ultimately dependent on
 * the individuals these relate and are removed if any of their associate
 * individuals are removed.
 *
 * A multiply-form is maximal if no two individuals combine into a single
 * individual under the respective operational behavior, and if no individuals
 * equal to nil.
 * 
 * The MultiplyForm class extends on the {Form} class.
 * It implements the form container as an ordered {TreeSet} and maintains
 * for each form a reference to the attribute sort to which it belongs.
 * It provides common implementations for some interface methods.
"""
class multiplyForm(form):
    """
    initialize the form
    self.elements is a set that contains all the elements (individuals or forms) 
    that belong to this form
    """
    def __init__(self, sort, *args, **kwargs):
        super(multiplyForm, self).__init__(sort)
        self.elements = []
        self.maximal = True
        
    def __iter__(self):
        return iter(self.elements)
        
    def __len__(self):
        return len(self.elements)
        
    """
    def __str__(self):
        s = '{'
        printComma = False
        for i in self.elements:
            if printComma:
                s += ', '
            else:
                printComma = True
            if 'ind' in i.type:
                if i.sort.type == 'sort-attribute' and i.hasAttribute():
                    s += str(i) + str(i.attribute)
                else:
                    s += str(i)
            else:
                s += str(i)
        s += '}'
        return s
    """
        
    def toString(self, string = '', printKey = False):
        if len(self) == 0:
            return '{}'
        if len(self) == 1:
            s = '{'
            s += self.elements[0].toString(printKey)
            if self.elements[0].sort.type == 'sort-attribute' and self.elements[0].hasAttribute():
                return s+self.elements[0].attribute.oneLineString(printKey)+'}'
            return s+'}'
        string += '  '
        s = '{\n' + string
        printComma = False
        for i in self.elements:
            if printComma:
                s += ',\n' + string
            else:
                printComma = True
            if 'ind' in i.type:
                s += i.toString(printKey)
                if i.sort.type == 'sort-attribute' and i.hasAttribute():
                    s += i.attribute.oneLineString(printKey)
            else:
                s += i.toString(string, printKey)
        s += '\n'+ string[:-2]+ '}'
        return s
        
    def oneLineString(self, printKey = False):
        s = '{'
        printComma = False
        for i in self.elements:
            if printComma:
                s += ', '
            else:
                printComma = True
            if 'ind' in i.type:
                s += i.toString(printKey)
                if i.sort.type == 'sort-attribute' and i.hasAttribute():
                    s += i.attribute.oneLineString()
            else:
                s += str(i)
        s += '}'
        return s
        
        """
    def printRhino(self, dX = 0, dY = 0):
        #print(string + '|' + self.sort.name+':',len(self.elements),'elements')
        for ind in self.elements:
            print(ind.printRhino(dX, dY))
            if ind.hasAttribute():
                ind.attribute.printRhino(dX, dY)
        return
        """
        
    def rhinoString(self, dX = 0, dY = 0):
        #print(string + '|' + self.sort.name+':',len(self.elements),'elements')
        s = ''
        for ind in self.elements:
            s += ind.printRhino(dX, dY) +'\n'
            if ind.hasAttribute():
                s += ind.attribute.rhinoString(dX, dY)
        return s
        
    def map(self, assignment, other):
        if self.sort != other.sort:
            return False
        count = 0
        for i in self.elements:
            if i.map(assignment, other):
                count+=1
        if count > 0:
            return True
        return False 
        
    """
    check if the form is empty
    """
    def isEmpty(self):
        return len(self.elements) == 0

    def totalLength(self):
        return len(self)
        
    def getElements(self):
        return self.elements
        
        
    def get(self, index):
        try:
            return self.elements[index]
        except IndexError:
            return None
            
    """
    return the first element of the form
    """
    def first(self):
        if len(self.elements)==0:
            return None
        else: return self.elements[0]
        
    """
    return the last element of the form
    """
    def last(self):
        if len(self.elements)==0:
            return None
        else: return self.elements[-1]
    
    def geometricFrame(self):
        minCoord = [None, None, None]
        maxCoord = [None, None, None]
        if len(self) == 0:
            return None
        for i in self:
            frame = i.geometricFrame()
            if frame is not None:
                for i in range(len(minCoord)):
                    if frame[0][i] is None:
                        continue
                    if minCoord[i] is None or minCoord[i] > frame[0][i]:
                        minCoord[i] = frame[0][i]
                for i in range(len(maxCoord)):
                    if frame[0][i] is None:
                        continue
                    if maxCoord[i] is None or maxCoord[i] < frame[1][i]:
                        maxCoord[i] = frame[1][i]
        return [minCoord, maxCoord]
                    
    """
    return the next element in the form after current element
    """
    def next(self, current):
        try:
            return self.elements[self.elements.index(current)+1]
        except:
            return None
        """
        if current is None:
            return None
        for i, e in enumerate(self.elements):
            if (e is current or e == current) and (i+1 < len(self.elements)):
                return self.elements[i+1]
        return None
        """
        
    """
    return the element that is equal/same as obj
    """
    def current(self, obj):
        for i, e in enumerate(self.elements):
            if e == obj and i+1 < len(self.elements):
                return self.elements[i]
        return None
        
    """
    return the previous element in the form before current element
    """
    def previous(self, current):
        try:
            return self.elements[self.elements.index(current)-1]
        except:
            return None
    
    """
    check if the form contains a certain element
    """
    def contains(self, current):
        for i, e in enumerate(self.elements):
            if e.__eq__(current):
                return True
        return False

    """
    check equality between two forms
    if the elements are the same between two forms, they are equal
    """
    def __eq__(self, other):
        if other is None or other.__class__ != self.__class__ or not other.sort == self.sort:
            return False
        if not self.maximal:
            self.maximalize()
        if not other.maximal:
            other.maximalize()
        if len(self) != len(other):
            return False
        for i in self.elements:
            check = False
            for j in other.elements:
                if i.type != j.type or i.sort != j.sort:
                    continue
                if i == j:
                    if 'ind' in i.type and i.sort.type == 'sort-attribute':
                        if i.attribute == j.attribute:
                            check = True
                    else:
                        check = True
            if check == False:
                return False
        return True
        
    """
    merging another form into this form
    i.e. inserting all elements belonging to other form into this form
    """
    def merge(self, other):
        for i in other.elements:
            self.insert(i)
            
    def move(self, other, append = False):
        if append:
            for i in self.elements:
                other.elements.append(i)
                i.assoc = other
            self.clear()
        else:
            for i in self.elements:
                other.insert(i)
            self.clear()
            
    """
    add an element into this form at a certain index
    """
    def add(self, index, element):
        if element.assoc is None:
            self.elements.insert(index, element)
            element.assoc = self
        else:
            newCopy = element.duplicate() 
            self.elements.insert(index, newCopy)
            newCopy.assoc = self
        self.maximal = False
        
    def duplicateForm(self):
        duplicate = self.sort.newForm()
        for i in self.elements:
            duplicate.insert(i.duplicate())
        return duplicate
        
    """
    remove an element from this form
    """
    def remove(self, element):
        try:
            self.elements.remove(element)
            element.assoc = None
        except ValueError:
            for i in self.elements:
                if i is element or i == element:
                    self.elements.remove(i)
            element.assoc = None
            
    """
    pop an element with index out of the list and return it
    index = 0 to pop the 1st element ... -1 for last element
    """
    def pop(self, index):
        return self.elements.pop(index)
        
    """
    going through this form for rule analysis or rule detection
    """
    def accept(self, visitor, direction=-1):
        if not self.isEmpty() and visitor.visitEnter(self, direction):
            if self.sort.type == 'sort-attribute':
                for element in self.elements:
                    if element.hasAttribute():
                        element.accept(visitor, direction)
                for element in self.elements:
                    if not element.hasAttribute():
                        element.accept(visitor, direction)
            else:
                for element in self.elements:
                    element.accept(visitor, direction)
        visitor.visitLeave(self)
        return
        
    """
    editing the form in a certain procedure(editor)
    mainly use for transforming the form elements with transformation matrix(es)
    """        
    def edit(self, editor):
        self.maximal = False
        result = self.sort.newForm()
        for i in self.elements:
            result.insert(i.edit(editor))
        return result
        
    """
    converting all individuals inside this form that uses vectors with coordinates into 
    using vectors with float values 
    """
    def convertFloat(self):
        result = self.sort.newForm()
        for i in self.elements:
            result.insert(i.convertFloat())
        result.maximal = self.maximal
        return result
        
    def splitElements(self):
        splits = []
        for i in self.elements:
            newForm = i.sort.newForm()
            newForm.insert(i)
            splits.append(newForm)
        return splits
                
    """
    check if this form contains a certain individual
    """
    def containsInd(self, ind):
        if self.type == 'form-metaForm':
            for i in self.elements:
                if i.sort == ind.sort or i.type == 'form-metaForm':
                    if i.containsInd(ind):
                        return True
        elif self.sort == ind.sort:
            for i in self.elements:
                if i.containsInd(ind):
                    return True
        return False
        
    """
    check if this form is contained by the compareForm after transforming
    with transformSet
    """
    def checkTransform(self, transformSet, compareForm):
        for i in self.elements:
            if not i.checkTransform(transformSet, compareForm):
                return False
        return True      
    
    """
    clear the content of the form
    """
    def clear(self):
        for i in self.elements:
            i.assoc = None
        self.elements = []
        
    #------------------------------------------------------------------
    
"""    
    def hashcode(self):
        return
    def toString(self):
        return
    def parse(self):
        return 
"""