# -*- coding: utf-8 -*-
"""
Created on Wed Mar 29 14:14:30 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
from builtins import range
from builtins import super
from builtins import str

from .plane import plane2D
from .circle import circle2D, circle3D

from ..io.parseReader import ParsingError
from ..transform.affineTransform import affineTransform2D, affineTransform3D
from ..struct.vector import vector2D, vector3D
from ..struct.coordinate import coordinate, c1, c2,  pi, acos, sqrt, cos

class circularArc2D(circle2D):
    
    type = 'ind-circularArc2D'
    
    def __init__(self, sort, center, start, end, pin, *args, **kwargs):
        if start == end:
            if pin is True:
                raise ValueError('For full circle arc, pin cannot be calculated')
            elif not start.colinear(center, pin):
                super(circularArc2D, self).__init__(sort, center, start)
                self.end = self.start = start
                self.pin = pin
                self.fullCircle = True
                return
            else:
                raise ValueError('Cannot create an arc with start point, end point being the same and pin is colinear with both')
        
        if pin is True:
            pin = plane2D.findPin(center, start, end)
            
        if plane2D.angleContainsPin(center, start, pin, end) == 1:
            start, end = end, start
            
        super(circularArc2D, self).__init__(sort, center, start)
        
        self.start = start
        self.end = end
        
        self.pin = pin
        self.fullCircle = False
        

    def angle(self):
        if self.fullCircle:
            return coordinate(2*pi)
        nom = self.start.dotproduct(self.end)
        denom = self.start.dotproduct(self.start) * self.end.dotproduct(self.end)
        compAngl = plane2D.compareAngle(self.center, self.start, self.end)
        if compAngl == -1:
            return coordinate(acos(nom.value / sqrt(denom.value)))
        elif compAngl == 1:
            return coordinate(2 * pi - acos(nom.value / sqrt(denom.value)))
        return coordinate(pi * (not self.start == self.end))
    
    def toString(self, printKey):
        s = ''
        if printKey and self.key is not None:
            s += '#'+str(self.key)+' '
        return s + '<'+str(self.center)+','+str(self.start)+','+str(self.end)+','+str(self.pin)+'>'

    def printRhino(self, dX = 0, dY = 0):
        return '_Arc '+str(self.center.x + dX)+','+str(self.center.y + dY)+'  '+str(self.end.x + dX)+','+str(self.end.y + dY)+'  '+str(self.angle()*180/pi)

    def printVal(self):
        #return str(self.start)+','+str(self.end)
        return '<c:'+str(self.center)+', s:'+str(self.start)+', e:'+str(self.end)+'>'
        return 'center:'+str(self.center)+', start:'+str(self.start)+', end:'+str(self.end)

    def equals(self, other):
        return self.center == other.center and self.start == other.start and self.end == other.end

    def isNil(self):
        if self.end is None or self.start is None:
            return True
        return self.start == self.end and not self.fullCircle
        
    def compareInterval(self, other):
        c = self.compareValue(other) 
        if c != 0:
            return c
        c = plane2D.compareRelAngle(self.center, self.center.add(vector2D(1,0)), self.start, other.start)
        if c != 0:
            return c
        return plane2D.compareRelAngle(self.center, self.start, self.end, other.end)
                
    def containsPoint(self, point):
        if not super(circularArc2D, self).containsPoint(point):
            return False
        return plane2D.angleContainsPin(self.center, self.start, point, self.end) != 1
        
    def compareEnd(self, other):
        return plane2D.angleContainsPin(self.center, self.start, self.end, other.end)
            
    def length(self):
        return self.radius * self.angle()
        
    def touches(self,other):
        if not self.compareValue(other) == 0:
            return False
        if self.fullCircle or other.fullCircle:
            return False
        return ((self.end == other.start and 
                 plane2D.angleContainsPin(self.center, self.start, other.end, self.end) != -1) or 
                (self.start == other.end and 
                 plane2D.angleContainsPin(self.center, self.start, other.start, self.end) != -1)) 
    
    def contains(self,other):
        if not self.compareValue(other) == 0:
            return False
        if self.fullCircle:
            return True
        if other.fullCircle:
            return False
        if self.start == other.start and self.end == other.end:
            return True
        selfContainsOtherStart = plane2D.angleContainsPin(self.center, self.start, other.start, self.end)
        selfContainsOtherEnd = plane2D.angleContainsPin(self.center, self.start, other.end, self.end)

        if (selfContainsOtherStart <= 0 and selfContainsOtherEnd <= 0 and 
            plane2D.angleContainsPin(self.center, other.start, self.start, other.end) <= 0 and 
            plane2D.angleContainsPin(self.center, other.start, self.end, other.end) <= 0):
                return False
        return (selfContainsOtherStart <= 0 and selfContainsOtherEnd <= 0 )
    
    def disjoint(self,other):
        if not self.compareValue(other) == 0:
            return False
        if self.fullCircle or other.fullCircle:
            return False
        return (plane2D.angleContainsPin(self.center, self.start, other.start, self.end) == 1 and 
                plane2D.angleContainsPin(self.center, self.start, other.end, self.end)  == 1 and 
                plane2D.angleContainsPin(self.center, other.start, self.start, other.end) == 1 and 
                plane2D.angleContainsPin(self.center, other.start, self.end, other.end)  == 1 )
    
    def align(self, other):
        if not self.compareValue(other) == 0:
            return False
        return self.end == other.end or self.start == other.start
        
    #@profile
    def common(self,other):
        result = []
        if self.fullCircle:
            new = circularArc2D(self.sort, other.center, other.start, other.end, other.pin)
            if self.hasAttribute():
                new.addAttribute(self.attribute)
            result.append(new)
            return result
        if other.fullCircle:
            new = circularArc2D(self.sort, self.center, self.start, self.end, self.pin)
            if self.hasAttribute():
                new.addAttribute(self.attribute)
            result.append(new)
            return result
            
        self_oStart = plane2D.angleContainsPin(self.center, self.start, other.start, self.end)
        self_oEnd = plane2D.angleContainsPin(self.center, self.start, other.end, self.end)
        other_sStart = plane2D.angleContainsPin(self.center, other.start, self.start, other.end)
        other_sEnd = plane2D.angleContainsPin(self.center, other.start, self.end, other.end)
        
        if self_oStart == 1 and self_oEnd == 1 and other_sStart == 1 and other_sEnd == 1:
            return result
            
        if (self.end == other.start and self_oEnd != -1) or (self.start == other.end and self_oStart != -1):
            return result
        
        if self_oStart == -1 and self_oEnd == -1 and other_sStart == -1 and other_sEnd == -1:
            result.append(circularArc2D(self.sort, self.center, self.start, other.end, True))
            result.append(circularArc2D(self.sort, self.center, other.start , self.end, True))
            if self.hasAttribute():
                for i in result:
                    i.addAttribute(self.attribute)
            return result
            
        if self_oStart == -1:
            start = other.start
        else:
            start = self.start
        if self_oEnd == -1:
            end = other.end
        else:
            end = self.end
                
        result.append(circularArc2D(self.sort, self.center, start , end, True))
        if self.hasAttribute():
            result[0].addAttribute(self.attribute)
        return result

    #@profile
    def combine(self, other):
        if self.fullCircle:
            return False
        if other.fullCircle:
            self.start = other.start
            self.end = other.end
            self.pin = other.pin
            return True
            
        self_oStart = plane2D.angleContainsPin(self.center, self.start, other.start, self.end)
        self_oEnd = plane2D.angleContainsPin(self.center, self.start, other.end, self.end)
        other_sStart = plane2D.angleContainsPin(self.center, other.start, self.start, other.end)
        other_sEnd = plane2D.angleContainsPin(self.center, other.start, self.end, other.end)
        
        if self_oStart == 1 and self_oEnd == 1 and other_sStart == 1 and other_sEnd == 1:
            return False
            
        if self_oStart != 1 and self_oEnd != 1 and other_sStart != 1 and other_sEnd != 1:
            if self.start == other.start and self.end == other.end:
                return False
            if self.pin.colinear(self.start, self.center):
                self.pin = self.end
                if self.pin.colinear(self.start, self.center):
                    self.pin = plane2D.findPin(self.center, self.start, self.end)
            self.end = self.start
            self.fullCircle = True
            return True
            
        if other_sStart < 1:
            self.start = other.start
        if other_sEnd < 1:
            self.end = other.end
        self.pin = plane2D.findPin(self.center, self.start, self.end)
                        
        return True
        
    #@profile
    def complement(self,other):
        result = []
        if other.fullCircle:
            return result
        if self.fullCircle:
            result.append(other.duplicateInd())
            return result
            
        self_oStart = plane2D.angleContainsPin(self.center, self.start, other.start, self.end)
        self_oEnd = plane2D.angleContainsPin(self.center, self.start, other.end, self.end)
        other_sStart = plane2D.angleContainsPin(self.center, other.start, self.start, other.end)
        other_sEnd = plane2D.angleContainsPin(self.center, other.start, self.end, other.end)


        if self_oStart != 1 and self_oEnd != 1 and other_sStart != 1 and other_sEnd != 1:
            result.append(circularArc2D(self.sort, self.center, other.end, other.start, True))
            
        else:
            if self_oStart == -1:
                result.append(circularArc2D(self.sort, self.center, self.start, other.start, True))
            if self_oEnd == -1:
                result.append(circularArc2D(self.sort, self.center, other.end, self.end, True))
                
            
        if self.hasAttribute():
            for i in result:
                i.addAttribute(self.attribute)
        return result
        
    def intersect(self, other):
        raise NotImplementedError
        
    def midpoint(self):
        angle = self.angle()
        midLine = vector2D((self.start.x+self.end.x)/c2, (self.start.y+self.end.y)/c2)
        midPoint = midLine.subtract(self.center).scale(c1/cos(self.angle.value/2)).add(self.center)
        if angle > pi:
            return self.center.subtract(midPoint.subtract(self.center))
        return midPoint
        
    def replicateInd(self, sort = 0):
        if sort == 0:
            sort = self.sort
        return circularArc2D(sort, self.center, self.start, self.end, self.pin) 
        
    def codescriptor(self):
        cirle = circle2D(self.sort, self.center, self.start, self.end)
        cirle.assoc = self
        return cirle
        
    def splitArc(self):
        ref = self.center.add(vector2D(self.radius,0))
        if plane2D.angleContainsPin(self.center, self.start, ref, self.end) == -1:
            a = circularArc2D(self.sort, self.center, ref, self.end, True)
            b = circularArc2D(self.sort, self.center, self.start, ref, True)
            if self.hasAttribute():
                a.addAttribute(self.attribute)
                b.addAttribute(self.attribute)
            return [a,b]
        return None
        
    def crossingOver(self):
        return plane2D.angleContainsPin(self.center, self.start, 
                                       self.center.add(vector2D(1,0)), 
                                       self.end)
    
    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform2D)
        if affine is None:
            return self.duplicateInd()
        start = self.start.transform(affine)
        end = self.end.transform(affine)         
        pin = self.pin.transform(affine)
        center = self.center.transform(affine)    
        

        check = 1
        for i in range(len(affine.matrix)):
            if affine.matrix[i][i] < 0:
                check *= -1
        if check == -1:
            temp = end
            end = start         
            start = temp

        return circularArc2D(self.sort, center, start, end, pin)
                
    def checkTransform(self, transform, compareForm):
        affine = transform.retrieveTransform(affineTransform2D)
        if affine is None:
            tfArc = self
        else:
            start = self.start.transform(affine)
            end = self.end.transform(affine)         
            pin = self.pin.transform(affine)
            center = self.center.transform(affine)         
            tfArc = circularArc2D(self.sort, center, start, end, pin)
            if self.hasAttribute():
                tfArc.addAttribute(self.attribute)
                
        splits = tfArc.splitArc()
        if splits is None:
            if not compareForm.containsInd(tfArc):
                return False
        else:
            for arc in splits:
                if not compareForm.containsInd(arc):
                    return False
        return True
        
    def containsInd(self, ind):
        if not self.contains(ind):
            return False
        if self.hasAttribute():
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circular arc')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y = vector2D.readVector(reader)
        center = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        start = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        end = vector2D(x,y)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y = vector2D.readVector(reader)
        pin = vector2D(x,y)
            
        ind = circularArc2D(sort, center, start, end, pin)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line segment')
            
        reader.newToken() #next token
        
        return ind

class circularArc3D(circle3D):
    type = 'ind-circularArc3D'
    
    def __init__(self, sort, center, start, end, pin, *args, **kwargs):
        if start == end:
            if pin is True:
                raise ValueError('For full circle arc, pin cannot be calculated')
            elif not start.colinear(center, pin):
                super(circularArc3D, self).__init__(sort, center, start, pin)
                self.end = self.start = start
                self.localEnd = self.localStart = self.localize(start)
                self.pin = pin
                self.localPin = self.localize(pin)
                self.fullCircle = True
                return
            else:
                raise ValueError('Cannot create an arc with start point, end point being the same and pin is colinear with both')
        
        if not start.colinear(center, end):
            super(circularArc3D, self).__init__(sort, center, start, end)
        else:
            super(circularArc3D, self).__init__(sort, center, start, pin)
            self.end = end
            self.localEnd = self.localize(end)
            
        self.start = start
        self.end = end
        self.localStart = self.localize(start)
        self.localEnd = self.localize(end)
        
        self.pin = pin
        self.localPin = self.localize(pin)
        
        if plane2D.angleContainsPin(self.localCenter, self.localStart, self.localPin, self.localEnd) == 1:
            self.start, self.end, self.localStart, self.localEnd = self.end, self.start, self.localEnd, self.localStart
        
        self.fullCircle = False

    def angle(self):
        nom = self.localStart.dotproduct(self.localEnd)
        denom = self.localStart.dotproduct(self.localStart) * self.localEnd.dotproduct(self.localEnd)
        compAngl = plane2D.compareAngle(self.localCenter, self.localStart, self.localEnd)
        if compAngl == -1:
            return coordinate(acos(nom.value / sqrt(denom.value)))
        elif compAngl == 1:
            return coordinate(2 * pi - acos(nom.value / sqrt(denom.value)))
        return coordinate(pi * (not self.localStart == self.localEnd))
    
    def toString(self, printKey):
        s = ''
        if printKey and self.key is not None:
            s += '#'+str(self.key)+' '
        return s + '<'+str(self.center)+','+str(self.start)+','+str(self.end)+','+str(self.pin)+'>'
                
                
    def printRhino(self, dX = 0, dY = 0):
        c = self.center
        s = self.end
        e = self.angle()*180 / pi
        return '_Arc '+str(c.x + dX)+','+str(c.y + dY)+','+str(c.z)+' t '+str(s.x + dX)+','+str(s.y + dY)+','+str(s.z)+'  '+str(e)


    def printVal(self):
        #return str(self.start)+','+str(self.end)
        return ('<c:'+str(self.center)+', s:'+str(self.start)+', e:'+
                str(self.end)+'>')
        return ('center:'+str(self.center)+', start:'+str(self.start)+
                ', end:'+str(self.end))

    def equals(self, other):
        return self.localCenter == other.localCenter and self.localStart == other.localStart and self.localEnd == other.localEnd

    def isNil(self):
        if self.localEnd is None or self.localStart is None:
            return True
        return self.localStart == self.localEnd and not self.fullCircle
        
    def compareInterval(self, other):
        c = self.compareValue(other)
        if c != 0:
            return c
        c = plane2D.compareRelAngle(self.localCenter, self.localCenter.add(vector2D(1,0)), self.localStart, other.localStart)
        if c != 0:
            return c
        return plane2D.compareRelAngle(self.localCenter, self.localStart, self.localEnd, other.localEnd)
                
    def containsPoint(self, point):
        if not super(circularArc3D, self).containsPoint(point):
            return False
        return plane2D.angleContainsPin(self.localCenter, self.localStart, self.localize(point), self.localEnd) != 1
        
    def compareEnd(self, other):
        return plane2D.angleContainsPin(self.localCenter, self.localStart, self.localEnd, other.localEnd)
            
    def length(self):
        return self.radius * self.angle()
        
    def touches(self,other):
        if not self.compareValue(other) == 0:
            return False
        if self.fullCircle or other.fullCircle:
            return False
        return ((self.localEnd == other.localStart and 
                 plane2D.angleContainsPin(self.localCenter, self.localStart, other.localEnd, self.localEnd) != -1) or 
                (self.localStart == other.localEnd and 
                 plane2D.angleContainsPin(self.localCenter, self.localStart, other.localStart, self.localEnd) != -1)) 
    
    def contains(self,other):
        if not self.compareValue(other) == 0:
            return False
        if self.fullCircle:
            return True
        if other.fullCircle:
            return False
        if self.localStart == other.localStart and self.localEnd == other.localEnd:
            return True
        selfContainsOtherStart = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localStart, self.localEnd)
        selfContainsOtherEnd = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localEnd, self.localEnd)
        if (selfContainsOtherStart <= 0 and selfContainsOtherEnd <= 0 and 
            plane2D.angleContainsPin(self.localCenter, other.localStart, self.localStart, other.localEnd) <= 0 and 
            plane2D.angleContainsPin(self.localCenter, other.localStart, self.localEnd, other.localEnd) <= 0):
                return False
        return (selfContainsOtherStart <= 0 and selfContainsOtherEnd <= 0 )
    
    def disjoint(self,other):
        if not self.compareValue(other) == 0:
            return False
        if self.fullCircle or other.fullCircle:
            return False
        return (plane2D.angleContainsPin(self.localCenter, self.localStart, other.localStart, self.localEnd) == 1 and 
                plane2D.angleContainsPin(self.localCenter, self.localStart, other.localEnd, self.localEnd)  == 1 and 
                plane2D.angleContainsPin(self.localCenter, other.localStart, self.localStart, other.localEnd) == 1 and 
                plane2D.angleContainsPin(self.localCenter, other.localStart, self.localEnd, other.localEnd)  == 1 )
    
    def align(self, other):
        if not self.compareValue(other) == 0:
            return False
        return self.localEnd == other.localEnd or self.localStart == other.localStart
        
    def common(self,other):
        result = []
        if self.fullCircle:
            new = circularArc3D(self.sort, self.center, self.start, self.end, self.pin)
            if self.sort.type == 'sort-attribute':
                new.addAttribute(self.attribute)
            result.append(new)
            return result
        if other.fullCircle:
            new = circularArc3D(self.sort, self.center, self.start, self.end, self.pin)
            if self.sort.type == 'sort-attribute':
                new.addAttribute(self.attribute)
            result.append(new)
            return result
            
        self_oStart = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localStart, self.localEnd)
        self_oEnd = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localEnd, self.localEnd)
        other_sStart = plane2D.angleContainsPin(self.localCenter, other.localStart, self.localStart, other.localEnd)
        other_sEnd = plane2D.angleContainsPin(self.localCenter, other.localStart, self.localEnd, other.localEnd)
        
        if self_oStart == 1 and self_oEnd == 1 and other_sStart == 1 and other_sEnd == 1:
            return result
            
        if (self.localEnd == other.localStart and self_oEnd != -1) or (self.localStart == other.localEnd and self_oStart != -1):
            return result
        
        if self_oStart == -1 and self_oEnd == -1 and other_sStart == -1 and other_sEnd == -1:
            pin = self.position(plane2D.findPin(self.localCenter, self.localStart, other.localEnd))
            result.append(circularArc3D(self.sort, self.center, self.start,  self.end, pin))
            pin = self.position(plane2D.findPin(self.localCenter, other.localStart, self.localEnd))
            result.append(circularArc3D(self.sort, self.center,  self.start , self.end,  pin))
            if self.sort.type == 'sort-attribute':
                for i in result:
                    i.addAttribute(self.attribute)
            return result
            
        if self_oStart == -1:
            start = other.localStart
        else:
            start = self.localStart
        if self_oEnd == -1:
            end = other.localEnd
        else:
            end = self.localEnd
            
        pin = self.position(plane2D.findPin(self.localCenter, start, end))
        result.append(circularArc3D(self.sort, self.center, self.position(start), self.position(end), pin))
        if self.sort.type == 'sort-attribute':
            result[0].addAttribute(self.attribute)
        return result

    def combine(self, other):
        if self.fullCircle:
            return False
        if other.fullCircle:
            self.localStart = other.localStart
            self.localEnd = other.localEnd
            self.localPin = other.localPin
            self.start = other.start
            self.end = other.end
            self.pin = other.pin
            return True
            
        self_oStart = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localStart, self.localEnd)
        self_oEnd = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localEnd, self.localEnd)
        other_sStart = plane2D.angleContainsPin(self.localCenter, other.localStart, self.localStart, other.localEnd)
        other_sEnd = plane2D.angleContainsPin(self.localCenter, other.localStart, self.localEnd, other.localEnd)
        
        if self_oStart == 1 and self_oEnd == 1 and other_sStart == 1 and other_sEnd == 1:
            return False
            
        if self_oStart != 1 and self_oEnd != 1 and other_sStart != 1 and other_sEnd != 1:
            if self.localStart == other.localStart and self.localEnd == other.localEnd:
                return False
            if self.localPin.colinear(self.localStart, self.localCenter):
                self.localPin = self.localEnd
                if self.localPin.colinear(self.localStart, self.localCenter):
                    self.localPin = plane2D.findPin(self.localCenter, self.localStart, self.localEnd)
            self.pin = self.position(self.localPin)
            self.localEnd = self.localStart
            self.end = self.start
            self.fullCircle = True
            return True
            
        if other_sStart < 1:
            self.localStart = other.localStart
            self.start = other.start
        if other_sEnd < 1:
            self.localEnd = other.localEnd
            self.end = other.end
        self.localPin = plane2D.findPin(self.localCenter, self.localStart, self.localEnd)
        self.pin = self.position(self.localPin)
        return True
        
    def complement(self,other):
        result = []
        if other.fullCircle:
            return result
        if self.fullCircle:
            new = circularArc3D(self.sort, self.center, self.end, self.start, self.pin)
            if self.hasAttribute():
                new.addAttribute(self.attribute)            
            result.append(new)
            return result
            
        self_oStart = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localStart, self.localEnd)
        self_oEnd = plane2D.angleContainsPin(self.localCenter, self.localStart, other.localEnd, self.localEnd)
        other_sStart = plane2D.angleContainsPin(self.localCenter, other.localStart, self.localStart, other.localEnd)
        other_sEnd = plane2D.angleContainsPin(self.localCenter, other.localStart, self.localEnd, other.localEnd)


        if self_oStart != 1 and self_oEnd != 1 and other_sStart != 1 and other_sEnd != 1:
            pin = self.position(plane2D.findPin(self.localCenter, other.localEnd, other.localStart))
            new = circularArc3D(self.sort, self.center, self.end, self.start, pin)
            result.append(new)
            
        else:
            if self_oStart == -1:
                pin = self.position(plane2D.findPin(self.localCenter, self.localStart, other.localStart))
                new = circularArc3D(self.sort, self.center, self.start, self.start, pin)
                result.append(new)
            if self_oEnd == -1:
                pin = self.position(plane2D.findPin(self.localCenter, other.localEnd, self.localEnd))
                new = circularArc3D(self.sort, self.center, self.end, self.end, pin)
                result.append(new)
            
        if self.sort.type == 'sort-attribute':
            for i in result:
                i.addAttribute(self.attribute)
        return result
        
    def intersect(self, other):
        raise NotImplementedError
        
    def midpoint(self):
        angle = self.angle()
        midLine = vector3D((self.localStart.x+self.localEnd.x)/c2, (self.localStart.y+self.localEnd.y)/c2, (self.localStart.z+self.localEnd.z)/c2)
        midPoint = midLine.subtract(self.localCenter).scale(c1/cos(self.angle.value/2)).add(self.localCenter)
        if angle > pi:
            return self.localCenter.subtract(midPoint.subtract(self.localCenter))
        return midPoint
        
    def splitArc(self):
        ref = self.localCenter.add(vector2D(self.radius,0))
        if plane2D.angleContainsPin(self.localCenter, self.localStart, ref, self.localEnd) == -1:
            pin = self.position(plane2D.findPin(self.localCenter, ref, self.localEnd))
            a = circularArc3D(self.sort, self.center, self.position(ref), self.end, pin)
            pin = self.position(plane2D.findPin(self.localCenter, self.localStart, ref))
            b = circularArc3D(self.sort, self.center, self.start, self.position(ref), pin)
            if self.hasAttribute():
                a.addAttribute(self.attribute)
                b.addAttribute(self.attribute)
            return [a,b]
        return None
        
    def crossingOver(self):
        return plane2D.angleContainsPin(self.localCenter, self.localStart, 
                                       self.localCenter.add(vector2D(1,0)), 
                                       self.localEnd)
          
    def replicateInd(self, sort = 0):
        if sort == 0:
            sort = self.sort
        return circularArc3D(sort, self.center, self.start, self.end, self.pin) 
        
    def codescriptor(self):
        cirle = circle3D(self.sort, self.center, self.start, self.pin)
        cirle.assoc = self
        return cirle
        

    def transformInd(self, transformSet):
        affine = transformSet.retrieveTransform(affineTransform3D)
        if affine is None:
            return self.duplicateInd()
        start = self.start.transform(affine)
        end = self.end.transform(affine)         
        pin = self.pin.transform(affine)
        center = self.center.transform(affine)         
        return circularArc3D(self.sort, center, start, end, pin)
                
    def checkTransform(self, transform, compareForm):
        affine = transform.retrieveTransform(affineTransform3D)
        if affine is None:
            tfArc = self
        else:
            start = self.start.transform(affine)
            end = self.end.transform(affine)         
            pin = self.pin.transform(affine)
            center = self.center.transform(affine)      
            tfArc = circularArc3D(self.sort, center, start, end, pin)
            if self.hasAttribute():
                tfArc.addAttribute(self.attribute)
        splits = tfArc.splitArc()
        if splits is None:
            if not compareForm.containsInd(tfArc):
                return False
        else:
            for arc in splits:
                if not compareForm.containsInd(arc):
                    return False
        return True
        
    def containsInd(self, ind):
        if not self.contains(ind):
            return False
        if self.hasAttribute():
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
    @staticmethod
    def parse(reader, sort):
        if reader.token() != '<':
            raise ParsingError(reader,'expected a "<" at the beginning of creating a circular arc')
        if reader.newToken() != '(': #bracket
            raise ParsingError(reader,'expected a "(" ')
            
        x,y,z = vector3D.readVector(reader)
        center = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        start = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        end = vector3D(x,y,z)
            
        reader.newToken() # bracket (current token would be ',')
        
        x,y,z = vector3D.readVector(reader)
        pin = vector3D(x,y,z)
            
        ind = circularArc3D(sort, center, start, end, pin)

        if reader.token() != '>':
            raise ParsingError(reader, '">" expected to close the line segment')
            
        reader.newToken() #next token
        
        return ind
