# -*- coding: utf-8 -*-
"""
Created on Fri Aug  5 10:26:35 2016

@author: akibdpt
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import super
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from copy import deepcopy, copy

from ..element import element
from ..sort.attributeSort import attributeSort

"""
 * An individual is the representational unit of a {Sort}.
 * An individual specifies a data element with a single value;
 * individuals may be collected into a {Form}.
 * The value an individual may hold is dependent on the sort it belongs to.
 * Every {PrimitiveSort} is defined by its characteristic individual,
 * i.e., a representational abstraction of all individuals of the sort,
 * and by a description of the algebraic behavior of its forms. The algebraic
 * behavior of forms of individuals is derived from a partial order relationship
 * defined on the individuals and forms of a sort. This partial order
 * relationship defines the following operations and relations on individuals:
 * an individual may contain another individual, this is always the case
 * if they are equal; otherwise, two individuals may be disjoint; if two
 * individuals are not disjoint, they combine into another individual
 * that contains both individuals; two individuals that are not disjoint may
 * have a common part that is contained by both individuals; and an
 * individual may have a complement part wrt another individual, this is
 * always the case of the other individual does not contain this individual.
 * An individual may have an associate individual (under the object-attribute
 * relationship). This is the associate of the form this individual belongs to.
 * 
 * The Individual class extends on the Element class. It is an
 * abstract class specifying the common framework for all characteristic
 * individual classes. An instance of any characteristic individual class
 * defines an individual. Every individual belongs to a {PrimitiveSort};
 * the individual is necessarily an instance of the sort's characteristic
 * individual class. An individual is an Element with an optional
 * reference {Key} and an optional attribute Form.
 * The reference key serves to relate individuals. The optional attribute form
 * must be consistent with the sort this individual belongs to.
"""
class individual(element):
    """
    initialize individual
    create an attribute form if the individual's sort is an attribute sort
    """
    
    
    def __init__(self, sort = None, *args, **kwargs):
        super(individual, self).__init__(sort)
        if sort is not None and sort.type == 'sort-attribute':
            self.attribute = self.sort.weight.newForm()
            self.attribute.assoc = self
        else: 
            self.attribute = None
        self.key = None
        #attributes methods

    def __str__(self):
        return self.toString(False)
        
    def toString(self, printKey):
        if printKey and self.key is not None:
            return '#'+str(self.key)+' '+str(self.value)
        return str(self.value)
        
    def printE(self, string = ''):
        print(string, end = ' _ ')
        print(self.printVal())
        if self.hasAttribute():
            self.attribute.printE(string+'  ')
            
    def printRhino(self, dX = 0, dY = 0):
        return None
        
    def geometricFrame(self):
        return None
        
    """
    check if an individual has attribute
    """
    def hasAttribute(self):
        return self.attribute is not None and not self.attribute.isEmpty()
        
    """
    get the attribute of the individual
    """
    def getAttribute(self):
        if (self.attribute is None) and isinstance(self.sort, attributeSort):
            print('ok')
            self.attribute = self.sort.weight.newForm()
            self.attribute.assoc = self
        else:
            print('not ok')
        return self.attribute
        
    """
    add an element as attribute to this individual
    """
    def addAttribute(self, element):
        if 'form' in element.type:
            for i in element:
                if i is None:
                    pass
                elif 'form' in i.type:
                    self.addAttribute(i)
                elif not i.isNil():
                    self.attribute.insert(i)
        elif 'ind' in element.type:
            self.attribute.insert(element)
        
    """
    not in use
    """
    def setAttribute(self, element):
        raise NotImplementedError
        if self.sort.type == 'sort-primitive':
            if self.assoc is not None:
                print("Attribute form cannot be set if individual belongs to a form")
                return
            #
        else: 
            #if isinstance(element.sort, recursiveSort):
            #   sort = sort.instance()
            if not element.sort == self.sort.weight:
                print("Attribute of erroneous sort")
                return
        if isinstance(element, individual):
            self.attribute = self.sort.newForm()
            self.attribute.add
        element.assoc = self
        self.attribute = element
        return 

    """
    check equality
    """
    def equals(self, other):
        if not self.value == other.value:
            return False
        if self.sort.type == 'sort-attribute':
            return self.attribute == other.attribute
        else:
            return True
        
    def isNil(self):
        return self.value is None
        
    def map(self, assignment, other):
        count = 0
        if self.assoc.hasAttribute():
            for i in self.assoc.attribute:
                if i.map(assignment, other):
                    count+=1
        if count > 0:
            return True
        return False 
    
    def compareValue(self, other):
        raise NotImplementedError 
        
    def duplicateInd(self, sort = None):
        duplicate =self.replicateInd()
        if sort is not None:
            duplicate.sort = sort
        if self.hasAttribute():
            duplicate.addAttribute(self.attribute)
        duplicate.key = self.key
        return duplicate
        
    def replicateInd(self):
        print(self.__class__)
        raise NotImplementedError 
        
    def updateAssoc(self):
        if self.hasAttribute():
            self.attribute.assoc = self
            self.attribute.updateAssoc()
        
    def setMin(self):
        raise NotImplementedError 
        
    """
    check the individual for rule analysis and rule detection
    """
    def accept(self, visitor, direction = -1):
        if self.isNil():
            return
        #if direction > 0:
        #    raise KeyError('is this even used?')
        #    if visitor.visitAssoc(self) and self.assoc is not None:
        #        self.assoc.accept(visitor, direction)
        #elif direction < 0:
        visitor.visitEnter(self, direction)
        if self.hasAttribute():
            self.attribute.accept(visitor, direction)
        #else:
        #    visitor.visitAssoc(self)
        visitor.visitLeave(self)
    
    """
    return the codescriptor of the individual, usually the individual itself 
    except for line segment which has the line containing it as codescriptor
    """
    def codescriptor(self) :
        codesc = self.duplicateInd()
        codesc.assoc = self
        return codesc
    
    """
    editing the individual with a certain transformation (editor)
    """
    def edit(self, editor):
        if self.isNil():
            return self.duplicateInd()
        if editor is not None:
            result = self.editInd(editor)
            if self.hasAttribute():
                result.sort = self.sort
                #result.attribute = result.sort.weight.newForm()
                att = self.attribute.edit(editor)
                result.addAttribute(att)
        else:
            result = self.duplicateInd()
            if self.hasAttribute():
                attr = self.attribute.edit(None)
                attr.assoc = result
                result.addAttribute(attr)
        if self.key is not None:
            result.key = self.key
        return result
        
    def getValue(self):
        return self.value
        
    """
    editing the individual with a certain transformation (editor)
    """
    def editInd(self, editor):
        if editor.type == 'transformSet':
            newInd = self.transformInd(editor)
            newInd.key = None
            return newInd
        elif editor.type == 'paramSet':
            return self.evaluateInd(editor)
        return self.replicateInd()
        
    def transformInd(self, transformSet):
        return copy(self)
        
    def evaluateInd(self, paramSet):
        return self.replicateInd()
        
    """
    converting the individuals that use coordinate vector into using float vectors  
    """
    def convertFloat(self):
        return copy(self)
            
    """
    check if the individual belongs to a certain form after transformation with
    a transform set
    """
    def checkTransform(self, transform, compareForm):
        comp = self.edit(transform)
        if compareForm.containsInd(comp):
            return True
        return False
        
    """
    check if this individual contains another individual
    """
    def containsInd(self, ind):
        if self != ind:
            return False
        if self.sort.type == 'sort-attribute':
            if not self.hasAttribute():
                if ind.hasAttribute():
                    return False
                else:
                    return True
            if ind.hasAttribute():
                if ind.attribute.checkPartOf(self.attribute):
                    return True
                else:
                    return False
            else:
                return True
        return True
        
    def clear(self):
        if self.attribute is not None:
            self.attribute.clear()
            
    @staticmethod
    def parse(reader, sort):
        raise NotImplementedError

    
